try:
    import jinja2  # noqa
    import openai  # noqa
    import tiktoken  # noqa
except ImportError:
    raise ImportError(  # noqa
        "not_again_ai.llm requires the 'llm' extra to be installed. "
        "You can install it using 'pip install not_again_ai[llm]'."
    )



import copy

from not_again_ai.llm.tokens import num_tokens_from_messages, truncate_str


def _inject_variable(
    messages_unformatted: list[dict[str, str]], variable_name: str, variable_text: str
) -> list[dict[str, str]]:
    """Injects variables into the messages using Python string formatting."""
    messages_formatted = copy.deepcopy(messages_unformatted)
    for message in messages_formatted:
        message["content"] = message["content"].replace("{{" + variable_name + "}}", variable_text)
    return messages_formatted


def priority_truncation(
    messages_unformatted: list[dict[str, str]],
    variables: dict[str, str],
    priority: list[str],
    token_limit: int,
    model: str = "gpt-3.5-turbo-0613",
) -> list[dict[str, str]]:
    """Formats messages_unformatted and injects variables into the messages in the order of priority, truncating the messages to fit the token limit.

    Algorithm:
        0. Checks if all variables in the priority list are in the variables dict. If not, adds the missing variables into priority in any order.
        1. Iterating over priority:
            a. Count the current number of tokens in messages_formatted and compute how many tokens remain.
            b. Count the number of times the variable occurs in messages_formatted.
            c. Truncate the variable to fit the remaining token budget taking into account the number of times it occurs in the messages.
            d. Inject the variable text into messages_formatted.

    Args:
        messages_unformatted: A list of dictionaries where each dictionary
            represents a message. Each message must have 'role' and 'content'
            keys with string values, where content is a string with any number of occurances of {{variable_name}}.
        variables: A dictionary where each key-value pair represents a variable name and its value to inject.
        priority: A list of variable names in their order of priority.
        token_limit: The maximum number of tokens allowed in the messages.
        model: The model to use for tokenization. Defaults to "gpt-3.5-turbo-0613".
    """

    # Check if all variables in the priority list are in the variables dict.
    # If not, add the missing variables into priority in any order.
    for var in variables:
        if var not in priority:
            priority.append(var)

    messages_formatted = copy.deepcopy(messages_unformatted)
    for var in priority:
        # Count the current number of tokens in messages_formatted and compute a remaining token budget.
        num_tokens = num_tokens_from_messages(messages_formatted, model=model)
        remaining_tokens = token_limit - num_tokens
        if remaining_tokens <= 0:
            break

        # Count the number of times the variable occurs in messages_formatted (including within the same message).
        num_var_occurrences = 0
        for message in messages_formatted:
            num_var_occurrences += message["content"].count("{{" + var + "}}")

        # Truncate the variable to fit the remaining token budget taking into account the number of times it occurs in the messages.
        truncated_var = truncate_str(variables[var], remaining_tokens // num_var_occurrences, model=model)

        # Inject the variable text into messages_formatted.
        messages_formatted = _inject_variable(messages_formatted, var, truncated_var)

    return messages_formatted



import json
from typing import Any

from openai import OpenAI


def chat_completion(
    messages: list[dict[str, str]],
    model: str,
    client: OpenAI,
    functions: list[dict[str, Any]] | None = None,
    max_tokens: int | None = None,
    temperature: float = 0.7,
    **kwargs: Any,
) -> dict[str, Any]:
    """Get an OpenAI chat completion response: https://platform.openai.com/docs/api-reference/chat/create

    Args:
        messages (list): A list of messages comprising the conversation so far.
        model (str): ID of the model to use. See the model endpoint compatibility table:
            https://platform.openai.com/docs/models/model-endpoint-compatibility
            for details on which models work with the Chat API.
        client (OpenAI): An instance of the OpenAI client.
        functions (list, optional): A list of functions the model may generate JSON inputs for. Defaults to None.
        max_tokens (int, optional): The maximum number of tokens to generate in the chat completion.
            Defaults to limited to the model's context length.
        temperature (float, optional): What sampling temperature to use, between 0 and 2.
            Higher values like 0.8 will make the output more random,
            while lower values like 0.2 will make it more focused and deterministic. Defaults to 0.7.
        **kwargs: Additional keyword arguments to pass to the OpenAI client chat completion.

    Returns:
        dict: A dictionary containing the following keys:
            - "finish_reason" (str): The reason the model stopped generating further tokens. Can be "stop" or "function_call".
            - "function_name" (str, optional): The name of the function called by the model, present only if "finish_reason" is "function_call".
            - "function_args" (dict, optional): The arguments of the function called by the model, present only if "finish_reason" is "function_call".
            - "message" (str, optional): The content of the generated assistant message, present only if "finish_reason" is "stop".
            - "completion_tokens" (int): The number of tokens used by the model to generate the completion.
            - "prompt_tokens" (int): The number of tokens in the generated response.
    """
    if functions is None:
        response = client.chat.completions.create(
            messages=messages,  # type: ignore
            model=model,
            max_tokens=max_tokens,
            temperature=temperature,
            n=1,
            **kwargs,
        )
    else:
        response = client.chat.completions.create(  # type: ignore
            messages=messages,
            model=model,
            functions=functions,
            function_call="auto",
            max_tokens=max_tokens,
            temperature=temperature,
            n=1,
            **kwargs,
        )

    response_data = {}
    finish_reason = response.choices[0].finish_reason
    response_data["finish_reason"] = finish_reason
    if finish_reason == "function_call":
        function_call = response.choices[0].message.function_call
        if function_call is not None:
            response_data["function_name"] = function_call.name  # type: ignore
            response_data["function_args"] = json.loads(function_call.arguments)
    elif finish_reason == "stop" or finish_reason == "length":
        message = response.choices[0].message
        response_data["message"] = message.content  # type: ignore
    usage = response.usage
    if usage is not None:
        response_data["completion_tokens"] = usage.completion_tokens  # type: ignore
        response_data["prompt_tokens"] = usage.prompt_tokens  # type: ignore
    return response_data



import tiktoken


def truncate_str(text: str, max_len: int, model: str = "gpt-3.5-turbo-0613") -> str:
    """Truncates a string to a maximum token length.

    Args:
        text: The string to truncate.
        max_len: The maximum number of tokens to keep.
        model: The model to use for tokenization. Defaults to "gpt-3.5-turbo-0613".
            See https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb
            for a list of OpenAI models.

    Returns:
        The truncated string.
    """
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        print("Warning: model not found. Using cl100k_base encoding.")
        encoding = tiktoken.get_encoding("cl100k_base")

    tokens = encoding.encode(text)
    if len(tokens) > max_len:
        tokens = tokens[:max_len]
        # Decode the tokens back to a string
        truncated_text = encoding.decode(tokens)
        return truncated_text
    else:
        return text


def num_tokens_in_string(text: str, model: str = "gpt-3.5-turbo-0613") -> int:
    """Return the number of tokens in a string.

    Args:
        text: The string to count the tokens.
        model: The model to use for tokenization. Defaults to "gpt-3.5-turbo-0613".

    Returns:
        The number of tokens in the string.
    """
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        print("Warning: model not found. Using cl100k_base encoding.")
        encoding = tiktoken.get_encoding("cl100k_base")
    return len(encoding.encode(text))


def num_tokens_from_messages(messages: list[dict[str, str]], model: str = "gpt-3.5-turbo-0613") -> int:
    """Return the number of tokens used by a list of messages.
    NOTE: Does not support counting tokens used by function calling.
    Reference: # https://github.com/openai/openai-cookbook/blob/main/examples/How_to_format_inputs_to_ChatGPT_models.ipynb

    Args:
        messages: A list of messages to count the tokens
            should ideally be the result after calling llm.prompts.chat_prompt.
        model: The model to use for tokenization. Defaults to "gpt-3.5-turbo-0613".
            See https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb
            for a list of OpenAI models.

    Returns:
        The number of tokens used by the messages.
    """
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        print("Warning: model not found. Using cl100k_base encoding.")
        encoding = tiktoken.get_encoding("cl100k_base")
    if model in {
        "gpt-3.5-turbo-0613",
        "gpt-3.5-turbo-16k-0613",
        "gpt-4-0314",
        "gpt-4-32k-0314",
        "gpt-4-0613",
        "gpt-4-32k-0613",
    }:
        tokens_per_message = 3  # every message follows <|start|>{role/name}\n{content}<|end|>\n
        tokens_per_name = 1  # if there's a name, the role is omitted
    elif model == "gpt-3.5-turbo-0301":
        tokens_per_message = 4
        tokens_per_name = -1
    elif "gpt-3.5-turbo" in model:
        return num_tokens_from_messages(messages, model="gpt-3.5-turbo-0613")
    elif "gpt-4" in model:
        return num_tokens_from_messages(messages, model="gpt-4-0613")
    else:
        raise NotImplementedError(
            f"""num_tokens_from_messages() is not implemented for model {model}.
See https://github.com/openai/openai-python/blob/main/chatml.md for information on how messages are converted to tokens."""
        )
    num_tokens = 0
    for message in messages:
        num_tokens += tokens_per_message
        for key, value in message.items():
            num_tokens += len(encoding.encode(value))
            if key == "name":
                num_tokens += tokens_per_name
    num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
    return num_tokens



import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import pandas as pd
import seaborn as sns

from not_again_ai.base.file_system import create_file_dir
from not_again_ai.viz.utils import reset_plot_libs


def scatterplot_basic(
    x: list[float] | (npt.NDArray[np.int_] | npt.NDArray[np.float_]),
    y: list[float] | (npt.NDArray[np.int_] | npt.NDArray[np.float_]),
    save_pathname: str,
    title: str | None = None,
    xlim: tuple[float, float] | None = None,
    ylim: tuple[float, float] | None = None,
    font_size: float = 48,
    height: float = 13,
    aspect: float = 1.2,
) -> None:
    """Saves a basic scatterplot to the specified pathname.

    Args:
        x (listlike): Input listlike data for x-axis
        y (listlike): Input listlike data for y-axis
        save_pathname (str): Filepath to save plot to. Parent directories will be automatically created.
        title (str, optional): Title of the plot. Defaults to None.
        xlim (tuple[float, float], optional): Set the x-axis limits (lower, upper). Defaults to None.
        ylim (tuple[float, float], optional): Set the y-axis limits (lower, upper). Defaults to None.
        font_size (float, optional): Font size. Defaults to 48.
        height (float, optional): Height (in inches) of the plot. Defaults to 13.
        aspect (float, optional): Aspect ratio of the plot, so that `aspect` * `height` gives the width of each facet in inches. Defaults to 1.2.
    """

    sns.set_theme(
        style="white",
        rc={
            "font.size": font_size,
            "axes.titlesize": font_size,
            "axes.labelsize": font_size * 0.8,
            "xtick.labelsize": font_size * 0.65,
            "ytick.labelsize": font_size * 0.65,
            "legend.fontsize": font_size * 0.5,
            "legend.title_fontsize": font_size * 0.55,
        },
    )
    data = pd.DataFrame({"x": x, "y": y})
    ax = sns.scatterplot(data=data, x="x", y="y")

    ax.figure.set_size_inches(height * aspect, height)
    ax.set_title(title)
    ax.set_xlabel("x")
    ax.set_ylabel("y")

    ax.set_xlim(xlim)
    ax.set_ylim(ylim)

    sns.despine(top=True, right=True)

    create_file_dir(save_pathname)
    plt.savefig(save_pathname, bbox_inches="tight")
    reset_plot_libs()



import jinja2


def _validate_message(message: dict[str, str]) -> bool:
    valid_fields = ["role", "content", "name", "function_call"]
    # Check if the only keys in the message are in valid_fields
    if not all(key in valid_fields for key in message):
        return False

    # Check if the each of the values are strings except function_call
    if not all(isinstance(value, str) for key, value in message.items() if key != "function_call"):
        return False

    valid_roles = ["system", "user", "assistant", "function"]
    if message["role"] not in valid_roles:
        return False

    return True


def chat_prompt(messages_unformatted: list[dict[str, str]], variables: dict[str, str]) -> list[dict[str, str]]:
    """Formats a list of messages for OpenAI's chat completion API using Jinja2 templating.

    The content of each message is treated as a Jinja2 template that is rendered
    with the provided variables.

    Args:
        messages_unformatted: A list of dictionaries where each dictionary
            represents a message. Each message must have 'role' and 'content'
            keys with string values, where content is a Jinja2 template.
        variables: A dictionary where each key-value pair represents a variable
            name and its value for template rendering.

    Returns:
        A list of dictionaries with the same structure as `messages_unformatted`,
        but with the 'content' of each message with the provided `variables`.

    Examples:
        >>> messages = [
        ...     {"role": "system", "content": "You are a helpful assistant."},
        ...     {"role": "user", "content": "Help me {{task}}"}
        ... ]
        >>> vars = {"task": "write Python code for the fibonnaci sequence"}
        >>> chat_prompt(messages, vars)
        [
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": "Help me write Python code for the fibonnaci sequence"}
        ]
    """
    messages_formatted = messages_unformatted.copy()
    for message in messages_formatted:
        # Validate each message and return a ValueError if any message is invalid
        if not _validate_message(message):
            raise ValueError(f"Invalid message: {message}")
        message["content"] = jinja2.Template(message["content"]).render(**variables)
    return messages_formatted



import pathlib


def create_file_dir(filepath: str) -> None:
    """Creates the parent directories for the specified filepath.
    Does not throw any errors if the directories already exist.

    Args:
        filepath (str): path to a file
    """
    root_path = pathlib.Path(filepath).parent
    root_path.mkdir(parents=True, exist_ok=True)



from openai import OpenAI


class InvalidOAIAPITypeError(Exception):
    """Raised when an invalid OAIAPIType string is provided."""

    pass


def openai_client(
    api_type: str = "openai",
    api_key: str | None = None,
    organization: str | None = None,
    timeout: float | None = None,
    max_retries: int | None = None,
) -> OpenAI:
    """Create an OpenAI client instance based on the specified API type and other provided parameters.

    Args:
        api_type (str, optional): Type of the API to be used. Accepted values are 'openai' or 'azure_openai'.
            Defaults to 'openai'.
        api_key (str, optional): The API key to authenticate the client. If not provided,
            OpenAI automatically uses `OPENAI_API_KEY` from the environment.
        organization (str, optional): The ID of the organization (for enterprise users). If not provided,
            OpenAI automotically uses `OPENAI_ORG_ID` from the environment.
        timeout (float, optional): TBD
        max_retries (int, optional): TBD

    Returns:
        OpenAI: An instance of the OpenAI client.

    Raises:
        InvalidOAIAPITypeError: If an invalid API type string is provided.
        NotImplementedError: If the specified API type is recognized but not yet supported (e.g., 'azure_openai').

    Examples:
        >>> client = oai_client(api_type="openai", api_key="YOUR_API_KEY")
    """
    if api_type not in ["openai", "azure_openai"]:
        raise InvalidOAIAPITypeError(f"Invalid OAIAPIType: {api_type}. Must be 'openai' or 'azure_openai'.")

    if api_type == "openai":
        args = {
            "api_key": api_key,
            "organization": organization,
            "timeout": timeout,
            "max_retries": max_retries,
        }
        # Remove any None values in order to use the default values.
        filtered_args = {k: v for k, v in args.items() if v is not None}
        return OpenAI(**filtered_args)  # type: ignore
    elif api_type == "azure_openai":
        raise NotImplementedError("AzureOpenAI is not yet supported by not-again-ai.")
    else:
        raise NotImplementedError("This should never happen.")



try:
    import numpy  # noqa
    import pandas  # noqa
    import seaborn  # noqa
except ImportError:
    raise ImportError(  # noqa
        "not_again_ai.viz requires the 'viz' extra to be installed. "
        "You can install it using 'pip install not_again_ai[viz]'."
    )
